from abc import ABC, abstractmethod
from typing import List, Tuple

from super_scad.scad.Unit import Unit
from super_scad.type.Vector2 import Vector2

from super_scad_thread.enum.ThreadAnatomy import ThreadAnatomy


class ThreadProfileCreator(ABC):
    """
    Abstract parent class for creating thread profiles.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 pitch: float):
        """
        Object constructor.

        :param pitch: The pitch of the thread.
        """

        self.__pitch: float = pitch
        """
        The pitch of the thread.
        """

        self.__master_profile: List[Tuple[ThreadAnatomy, Vector2]] | \
                               Tuple[Tuple[ThreadAnatomy, Vector2], ...] | None = None
        """
        The 2D master profile of the thread. I.e. all the 2D points of the profile for one pitch.
        """

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def pitch(self) -> float:
        """
        Return the pitch of the thread.
        """
        return self.__pitch

    # ------------------------------------------------------------------------------------------------------------------
    @property
    @abstractmethod
    def unit_of_length(self) -> Unit:
        """
        Returns the unit of length in which the thread is specified.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def number_of_points_per_pitch(self) -> int:
        """
        Returns the number of points for one pitch.
        """
        if self.__master_profile is None:
            self.__master_profile = self.create_master_profile

        return len(self.__master_profile)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    @abstractmethod
    def minor_diameter(self) -> float:
        """
        Returns the minor diameter of the thread.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    @property
    @abstractmethod
    def major_diameter(self) -> float:
        """
        Returns the major diameter of the thread.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    @abstractmethod
    def create_master_profile(self) -> List[Tuple[ThreadAnatomy, Vector2]] | \
                                       Tuple[Tuple[ThreadAnatomy, Vector2], ...]:
        """
        Returns the 2D master profile of the thread. I.e., all the 2D points of the profile for one pitch. The thread
        profile must start in the middle of the root.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    def create_thread_profile(self, length: float) -> Tuple[List[ThreadAnatomy], List[Vector2]]:
        """
        Returns the thread profile points.
        """
        if self.__master_profile is None:
            self.__master_profile = self.create_master_profile()

        anatomy = []
        points = []

        if self.__master_profile[0][1].y != 0.0:
            anatomy.append(self.__master_profile[0][0])
            points.append(Vector2(self.__master_profile[0][1].x, -self.__pitch))

        for revolution in range(-1, int(length / self.__pitch) + 2):
            for j in range(0, len(self.__master_profile)):
                x = self.__master_profile[j][1].x
                y = self.__master_profile[j][1].y + revolution * self.__pitch
                if y <= (length + self.__pitch):
                    anatomy.append(self.__master_profile[j][0])
                    points.append(Vector2(x, y))

        return anatomy, points

# ----------------------------------------------------------------------------------------------------------------------



from abc import ABC

from super_scad_thread.lead_thread.ThreadLeadCreator import ThreadLeadCreator


class InternalThreadLeadCreator(ThreadLeadCreator, ABC):
    """
    Creates a lead thread on a 2D internal thread profile.
    """
    # Nothing to implement.
    pass

# ----------------------------------------------------------------------------------------------------------------------



import math
from typing import Any, Dict, List, Tuple

from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.type import Vector3

from super_scad_thread.enum.ThreadDirection import ThreadDirection
from super_scad_thread.lead_thread.external.ExternalThreadLeadCreator import ExternalThreadLeadCreator
from super_scad_thread.Thread import Thread
from super_scad_thread.ThreadProfileCreator import ThreadProfileCreator


class ExternalThread(Thread):
    """
    SuperSCAD widget for creating external threads.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 length: float,
                 thread_profile_creator: ThreadProfileCreator,
                 top_thread_lead_creator: ExternalThreadLeadCreator,
                 bottom_thread_lead_creator: ExternalThreadLeadCreator,
                 direction: ThreadDirection,
                 inner_radius: float | None = None,
                 inner_diameter: float | None = None,
                 center: bool = False,
                 convexity: int = 2):
        """
        Object contructor.

        :param length: The length of the thread.
        :param thread_profile_creator: The thread profile creator.
        :param top_thread_lead_creator: The object for creating a lead on the top of the thread.
        :param bottom_thread_lead_creator: The object for creating a lead on the top of the thread.
        :param direction: The direction of the thread.
        :param inner_radius: For a hollow external thread, the outer radius of the thread.
        :param inner_diameter: For a hollow external thread, the outer diameter of the thread.
        :param center: Whether to center the thread along the z-axis.
        :param convexity: The convexity of the thread. Normally 2 is enough, however, in some cases, a higher value is
                          required.
        """
        Thread.__init__(self,
                        length=length,
                        thread_profile_creator=thread_profile_creator,
                        top_thread_lead_creator=top_thread_lead_creator,
                        bottom_thread_lead_creator=bottom_thread_lead_creator,
                        direction=direction,
                        center=center,
                        convexity=convexity)

        self._inner_radius: float | None = inner_radius
        """
        For a hollow external thread, the outer radius of the thread.
        """

        self._inner_diameter: float | None = inner_diameter
        """
        For a hollow external thread, the outer diameter of the thread.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)
        validator.validate_exclusive({'inner_radius'}, {'inner_diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def inner_radius(self) -> float | None:
        """
        Returns the internal radius of the external thread. 
        """
        if self._inner_radius is not None:
            return self._inner_radius

        if self._inner_diameter is not None:
            return 0.5 * self._inner_diameter

        return None

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def inner_diameter(self) -> float | None:
        """
        Returns the internal diameter of the external thread. 
        """
        if self._inner_diameter is not None:
            return self._inner_diameter

        if self._inner_radius is not None:
            return 2.0 * self._inner_radius

        return None

    # ------------------------------------------------------------------------------------------------------------------
    def __create_faces_end_solid(self,
                                 faces: List[List[Vector3] | Tuple[Vector3, ...]],
                                 thread_3d: List[List[Vector3]]) -> None:
        """
        Creates faces for the top and bottom ends of the thread for a solid external thread.

        :param faces: The list of faces.
        :param thread_3d: The thread profile in 3D.
        """
        edges = len(thread_3d)
        number_of_points_per_pitch = self.thread_profile_creator.number_of_points_per_pitch

        # Add bottom face.
        bottom_face = []
        for edge in range(0, edges):
            bottom_face.append(thread_3d[edge][0])
        bottom_face.append(thread_3d[0][number_of_points_per_pitch])
        faces.append(bottom_face)

        # Add top face.
        top_face = []
        for index in range(edges - 1, 0, -1):
            top_face.append(thread_3d[index][len(thread_3d[index]) - 1])
        top_face.append(thread_3d[0][len(thread_3d[0]) - 1])
        faces.append(top_face)

    # ------------------------------------------------------------------------------------------------------------------
    def __create_faces_ends_hollow(self,
                                   faces: List[List[Vector3] | Tuple[Vector3, ...]],
                                   thread_3d: List[List[Vector3]]) -> None:
        """
        Creates faces for the top and bottom ends of the thread for a hollow external thread.

        :param faces: The list of faces.
        :param thread_3d: The thread profile in 3D.
        """
        edges = len(thread_3d)
        inner_radius = self.inner_radius

        bottom_points = []
        top_points = []
        for edge in range(0, edges):
            angle = edge * 360.0 / edges
            x = inner_radius * math.cos(math.radians(angle))
            y = inner_radius * math.sin(math.radians(angle))
            bottom_points.append(Vector3(x, y, 0.0))
            top_points.append(Vector3(x, y, self.length))

        # Add bottom face.
        for edge in range(1, edges):
            faces.append((thread_3d[edge - 1][0],
                          thread_3d[edge][0],
                          bottom_points[edge],
                          bottom_points[edge - 1]))
        faces.append((thread_3d[edges - 1][0],
                      thread_3d[0][0],
                      bottom_points[0],
                      bottom_points[edges - 1]))

        # Add top face.
        for edge in range(1, edges):
            faces.append((top_points[edge - 1],
                          top_points[edge],
                          thread_3d[edge][len(thread_3d[edge]) - 1],
                          thread_3d[edge - 1][len(thread_3d[edge - 1]) - 1]))
        faces.append((top_points[edges - 1],
                      top_points[0],
                      thread_3d[0][len(thread_3d[0]) - 1],
                      thread_3d[edges - 1][len(thread_3d[edges - 1]) - 1]))

        # Add inner faces.
        for edge in range(1, edges):
            faces.append((bottom_points[edge - 1],
                          bottom_points[edge],
                          top_points[edge],
                          top_points[edge - 1]))
        faces.append((bottom_points[edges - 1],
                      bottom_points[0],
                      top_points[0],
                      top_points[edges - 1]))

    # ------------------------------------------------------------------------------------------------------------------
    def __create_faces_thread(self,
                              faces: List[List[Vector3] | Tuple[Vector3, ...]],
                              thread_3d: List[List[Vector3]]) -> None:
        """
        Creates the faces for the thread profile.

        :param faces: The list of faces.
        :param thread_3d: The thread profile in 3D.
        """
        edges = len(thread_3d)
        number_of_points_per_pitch = self.thread_profile_creator.number_of_points_per_pitch

        # Faces between the consecutive edges.
        for edge in range(1, edges):
            for key in range(1, len(thread_3d[0])):
                faces.append((thread_3d[edge - 1][key - 1],
                              thread_3d[edge - 1][key],
                              thread_3d[edge][key],
                              thread_3d[edge][key - 1]))

        # Faces between the last and first edges.
        for key in range(1, len(thread_3d[0]) - number_of_points_per_pitch - 1):
            faces.append((thread_3d[edges - 1][key - 1],
                          thread_3d[edges - 1][key],
                          thread_3d[0][key + number_of_points_per_pitch],
                          thread_3d[0][key + number_of_points_per_pitch - 1]))

    # ------------------------------------------------------------------------------------------------------------------
    def _create_faces(self, thread_3d: List[List[Vector3]]) -> List[List[Vector3] | Tuple[Vector3, ...]]:
        """
        Creates faces given a thread profile in 3D.

        :param thread_3d: The thread profile in 3D.
        """
        faces: List[List[Vector3] | Tuple[Vector3, ...]] = []

        self.__create_faces_thread(faces, thread_3d)
        if self.inner_radius is None:
            self.__create_faces_end_solid(faces, thread_3d)
        else:
            self.__create_faces_ends_hollow(faces, thread_3d)

        return faces

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.type.Vector2 import Vector2

from super_scad_thread.enum.ThreadAnatomy import ThreadAnatomy
from super_scad_thread.lead_thread.external.ExternalThreadLeadCreator import ExternalThreadLeadCreator


class NoneExternalThreadLeadCreator(ExternalThreadLeadCreator):
    """
    Creates no lead on a 2D external thread profile.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, pitch: float, minor_diameter: float):
        """
        Object constructor.

        :param pitch: The pitch of the thread.
        :param minor_diameter: The minor diameter of the thread.
        """

        self.__pitch: float = pitch
        """
        The pitch of the thread.
        """

        self.__minor_diameter: float = minor_diameter
        """
        The minor diameter of the thread. 
        """

    # ------------------------------------------------------------------------------------------------------------------
    def create_lead(self,
                    thread_profile: List[Vector2],
                    thread_anatomy: List[ThreadAnatomy],
                    z: float,
                    angle: float) -> List[Vector2]:
        """
        Creates a lead on a 2D thread profile.

        :param thread_profile: The thread profile in 2D.
        :param thread_anatomy: The location of a points on the thread profile.
        :param z: The current translation in z-axis direction.
        :param angle: The angle of the current rotation.

        It is guaranteed that 0.0 <= z < pitch.
        It is guaranteed that 0.0 <= angle < 360.0.
        """
        for index, point in enumerate(thread_profile):
            if point.y < 0.0:
                thread_profile[index] = Vector2(self.__minor_diameter / 2.0, 0.0)
            else:
                break

        return thread_profile

# ----------------------------------------------------------------------------------------------------------------------



import math
from abc import ABC
from typing import Any, Dict, List, Tuple

from super_scad.d3.Polyhedron import Polyhedron
from super_scad.scad.ArgumentValidator import ArgumentValidator
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Mirror3D import Mirror3D
from super_scad.transformation.Translate3D import Translate3D
from super_scad.type.Angle import Angle
from super_scad.type.Vector2 import Vector2
from super_scad.type.Vector3 import Vector3
from super_scad.util.Radius2Sides4n import Radius2Sides4n

from super_scad_thread.enum.ThreadAnatomy import ThreadAnatomy
from super_scad_thread.enum.ThreadDirection import ThreadDirection
from super_scad_thread.lead_thread.ThreadLeadCreator import ThreadLeadCreator
from super_scad_thread.ThreadProfileCreator import ThreadProfileCreator


class Thread(ScadWidget, ABC):
    """
    Abstract SuperSCAD widget for creating threads.
    """

    def __init__(self,
                 *,
                 length: float,
                 thread_profile_creator: ThreadProfileCreator,
                 top_thread_lead_creator: ThreadLeadCreator,
                 bottom_thread_lead_creator: ThreadLeadCreator,
                 direction: ThreadDirection,
                 center: bool = False,
                 convexity: int = 2):
        """
        Object constructor.

        :param length: The length of the thread.
        :param thread_profile_creator: The thread profile creator.
        :param top_thread_lead_creator: The object for creating a lead on the top of the thread.
        :param bottom_thread_lead_creator: The object for creating a lead on the top of the thread.
        :param direction: The direction of the thread.
        :param center: Whether to center the thread along the z-axis.
        :param convexity: The convexity of the thread. Normally 2 is enough, however, in some cases, a higher value is
                          required.
        """
        ScadWidget.__init__(self)

        self._length: float = length
        """
        The length of the thread.
        """

        self._thread_profile_creator: ThreadProfileCreator = thread_profile_creator
        """
        The thread profile creator.
        """

        self._top_thread_lead_creator: ThreadLeadCreator = top_thread_lead_creator
        """
        The object for creating a lead on the top of the thread.
        """

        self._bottom_thread_lead_creator: ThreadLeadCreator = bottom_thread_lead_creator
        """
        The object for creating a lead on the top of the thread.
        """

        self._direction: ThreadDirection = direction
        """
        The direction of the thread.
        """

        self._center: bool = center
        """
        Whether to center the thread along the z-axis.
        """

        self._convexity: int = convexity
        """
        The convexity of the thread. Normally 2 is enough.
        """

        self.__validate_arguments(locals())

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __validate_arguments(args: Dict[str, Any]) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.

        :param args: The arguments supplied to the constructor.
        """
        validator = ArgumentValidator(args)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether to center the thread along the z-axis.
        """
        return self._center

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def length(self) -> float:
        """
        Returns the length of the thread.
        """
        return self._length

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def minor_diameter(self) -> float:
        """
        Returns the minor diameter of the thread.
        """
        return self.thread_profile_creator.minor_diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def major_diameter(self) -> float:
        """
        Returns the major diameter of the thread.
        """
        return self.thread_profile_creator.major_diameter

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def thread_profile_creator(self) -> ThreadProfileCreator:
        """
        Returns the thread profile creator.
        """
        return self._thread_profile_creator

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def top_thread_lead_creator(self) -> ThreadLeadCreator:
        """
        Returns the object for creating a lead on the top of the thread.
        """
        return self._top_thread_lead_creator

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def bottom_thread_lead_creator(self) -> ThreadLeadCreator:
        """
        Returns the object for creating a lead on the bottom of the thread.
        """
        return self._bottom_thread_lead_creator

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def direction(self) -> ThreadDirection:
        """
        Returns direction of the thread.
        """
        return self._direction

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int:
        """
        Returns the convexity of the thread.
        """
        return self._convexity

    # ------------------------------------------------------------------------------------------------------------------
    def _create_faces(self, thread_3d: List[List[Vector3]]) -> List[List[Vector3] | Tuple[Vector3, ...]]:
        """
        Creates faces given a thread profile in 3D.

        :param thread_3d: The thread profile in 3D.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __thread_translate(master_thread_profile_2d: List[Vector2], z: float) -> List[Vector2]:
        """
        Apply the translation in the z-axis direction due to rotation.

        :param master_thread_profile_2d: The master thread profile in 2D.
        :param z: The translation in z-axis direction.
        """
        thread_profile_2d = []
        for key, point in enumerate(master_thread_profile_2d):
            thread_profile_2d.append(Vector2(point.x, point.y + z))

        return thread_profile_2d

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __thread_move_point_to_zero(thread_profile_2d: List[Vector2]) -> List[Vector2]:
        """
        Moves a thread profile point just below the zero y-axis at y=0.0 without modifying the profile.

        :param thread_profile_2d: The master thread profile in 2D.
        """
        found = False
        index = 0
        for index in range(1, len(thread_profile_2d)):
            if thread_profile_2d[index - 1].y < 0.0 < thread_profile_2d[index].y:
                found = True
                break
            elif thread_profile_2d[index].y > 0.0:
                break

        if found:
            try:
                m = ((thread_profile_2d[index - 1].y - thread_profile_2d[index].y) /
                     (thread_profile_2d[index - 1].x - thread_profile_2d[index].x))
                x = thread_profile_2d[index - 1].x - thread_profile_2d[index - 1].y * m
                thread_profile_2d[index - 1] = Vector2(x, 0.0)
            except ZeroDivisionError:
                thread_profile_2d[index - 1] = Vector2(thread_profile_2d[index - 1].x, 0.0)

        return thread_profile_2d

    # ------------------------------------------------------------------------------------------------------------------
    def __thread_create_lead(self,
                             thread_profile: List[Vector2],
                             thread_anatomy: List[ThreadAnatomy],
                             z: float,
                             angle: float,
                             lead_creator: ThreadLeadCreator) -> List[Vector2]:
        """
        Applies the tread lead transformation on the thread profile in 2D.

        :param thread_profile: The thread profile in 2D.
        :param thread_anatomy: The location of a points on the thread profile.
        :param z: The current translation in z-axis direction.
        :param angle: The rotation in degrees.
        :param lead_creator: The thread lead creator.
        """
        thread_profile = lead_creator.create_lead(thread_profile, thread_anatomy, z, angle)

        self.__thread_assert_lead(thread_profile)

        return thread_profile

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __thread_assert_lead(thread_profile_2d: List[Vector2]) -> None:
        """
        Assets the thread after applying a lead in or our.

        :param thread_profile_2d: The master thread profile in 2D.
        """
        assert thread_profile_2d[0].y == 0.0, f'The first point of a thread profile must at y=0.0 zero' \
                                              ', got {thread_profile_2d[0]}'
        for point in thread_profile_2d:
            assert point.y >= 0.0, 'All points of a thread profile must be at y>=0.0'

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __thread_reverse(thread_profile_2d: List[Vector2], length: float) -> List[Vector2]:
        """
        Reverse the thread profile and align the (original) top at z=0.0.

        :param thread_profile_2d: The thread profile in 2D.
        :param length: The (final) length of the thread.
        """
        reversed_thread_profile = []
        for key, point in enumerate(reversed(thread_profile_2d)):
            reversed_thread_profile.append(Vector2(point.x, length - point.y))

        return reversed_thread_profile

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __thread_transform_to_3d(thread_profile_2d: List[Vector2], angle: float) -> List[Vector3]:
        """
        Apply the transformation form 2D profile to a 3D profile

        :param thread_profile_2d: The thread profile in 2D.
        :param angle: The rotation in degrees.
        """
        thread_profile_3d = []
        for point in thread_profile_2d:
            x = point.x * math.cos(math.radians(angle))
            y = point.x * math.sin(math.radians(angle))
            z = point.y
            thread_profile_3d.append(Vector3(x, y, z))

        return thread_profile_3d

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        thread_profile_creator = self.thread_profile_creator
        context.set_unit_length_current(thread_profile_creator.unit_of_length)

        length = self.length
        master_thread_anatomy, master_thread_profile_2d = thread_profile_creator.create_thread_profile(length)
        master_thread_anatomy_reversed = list(reversed(master_thread_anatomy))
        pitch = thread_profile_creator.pitch

        top_thread_lead_creator = self.top_thread_lead_creator
        bottom_thread_lead_creator = self.bottom_thread_lead_creator

        # Offsets for the top end due to inversion of the thread.
        angle_offset = 360.0 * math.remainder(length, pitch)
        z_offset = pitch * angle_offset / 360.0

        thread_3d = []
        sides = Radius2Sides4n.r2sides4n(context, thread_profile_creator.major_diameter / 2.0)
        for i in range(0, sides):
            z = i * pitch / sides
            angle = i * 360.0 / sides
            thread_profile_2d = self.__thread_translate(master_thread_profile_2d, z)
            thread_profile_2d = self.__thread_move_point_to_zero(thread_profile_2d)
            thread_profile_2d = self.__thread_create_lead(thread_profile_2d,
                                                          master_thread_anatomy,
                                                          z,
                                                          angle,
                                                          bottom_thread_lead_creator)
            thread_profile_2d = self.__thread_reverse(thread_profile_2d, length)
            thread_profile_2d = self.__thread_move_point_to_zero(thread_profile_2d)
            thread_profile_2d = self.__thread_create_lead(thread_profile_2d,
                                                          master_thread_anatomy_reversed,
                                                          Angle.normalize(z_offset - z, pitch),
                                                          Angle.normalize(angle_offset - angle, 360.0),
                                                          top_thread_lead_creator)
            thread_profile_2d = self.__thread_reverse(thread_profile_2d, length)
            thread_profile_3d = self.__thread_transform_to_3d(thread_profile_2d, angle)
            thread_3d.append(thread_profile_3d)

        faces = self._create_faces(thread_3d)
        polyhedron = Polyhedron(faces=faces, highlight_faces=None, highlight_diameter=0.1, convexity=self.convexity)

        if self.direction == ThreadDirection.LEFT:
            polyhedron = Mirror3D(x=1.0, child=polyhedron)

        if self.center:
            polyhedron = Translate3D(z=-self.length / 2.0, child=polyhedron)

        return polyhedron

# ----------------------------------------------------------------------------------------------------------------------



from enum import auto, Enum, STRICT


class ThreadAnatomy(Enum, boundary=STRICT):
    """
    Enumeration of all possible locations of a point on a thread profile.
    """
    # ------------------------------------------------------------------------------------------------------------------
    AT_MINOR = auto()
    """
    For external thread the point is at the root.
    For internal thread the point is at the crest.
    """

    AT_MAJOR = auto()
    """
    For external thread the point is at the crest.
    For internal thread the point is at the root.
    """

    AT_FLANK = auto()
    """
    The point is on a flank of the thread.
    """

# ----------------------------------------------------------------------------------------------------------------------



from abc import ABC
from typing import List

from super_scad.type.Vector2 import Vector2

from super_scad_thread.enum.ThreadAnatomy import ThreadAnatomy


class ThreadLeadCreator(ABC):
    """
    Creates a lead thread on a 2D thread profile.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def create_lead(self,
                    thread_profile: List[Vector2],
                    thread_anatomy: List[ThreadAnatomy],
                    z: float,
                    angle: float) -> List[Vector2]:
        """
        Creates a lead thread on a 2D thread profile.

        :param thread_profile: The thread profile in 2D.
        :param thread_anatomy: The location of a points on the thread profile.
        :param z: The current translation in z-axis direction.
        :param angle: The angle of the current rotation.

        It is guaranteed that 0.0 <= z < pitch.
        It is guaranteed that 0.0 <= angle < 360.0.
        """
        raise NotImplementedError()

# ----------------------------------------------------------------------------------------------------------------------
